#ifndef _FFDB_H_
#define _FFDB_H_
#include <stddef.h>
#include <string>
#include <vector>

class Row {
	void adjustptrs(void *oldptr, void *newptr);
public:
	struct VL {
		void *ptr;
		size_t size;
		VL() {}
		VL(void *ptr_, size_t size_) : ptr(ptr_), size(size_) {}
	};
	typedef VL                                      value_type;
	typedef std::vector<value_type>                 container_type;
	typedef container_type::const_pointer           const_pointer;
	typedef container_type::const_reference         const_reference;
	typedef container_type::const_iterator          const_iterator;
	typedef container_type::const_reverse_iterator  const_reverse_iterator;
	typedef container_type::size_type               size_type;
	typedef container_type::difference_type         difference_type;
private:
	container_type col_ptr;
	size_t buffer_bytes;
	size_t buffer_size;
	void *buffer;
public:
	
	Row();
	~Row() { clear(); }
	Row(const Row &row_);
	Row& operator=(const Row &row_);
	
	void reserve_raw(size_t bytes);
	void reserve_columns(size_type columns)         { col_ptr.reserve(columns); }
	void *alloc(size_t bytes, size_t alignment);
	void *alloc(size_t bytes, size_t alignment, const void *value);
	void resize_columns(size_type columns);
	
	bool operator==(const Row &row2) const;
	bool operator!=(const Row &row2) const;
	
	size_t raw_size() const                         { return buffer_bytes; }
	size_type size() const                          { return col_ptr.size(); }
	bool empty() const                              { return col_ptr.empty(); }
	void clear();
	
	void swap(Row &row2);
	
	const_iterator begin() const                    { return col_ptr.begin(); }
	const_iterator end() const                      { return col_ptr.end(); }
	const_reverse_iterator rbegin() const           { return col_ptr.rbegin(); }
	const_reverse_iterator rend() const             { return col_ptr.rend(); }
	const_reference operator[](size_type idx) const { return col_ptr[idx]; }
	const_reference front() const                   { return col_ptr.front(); }
	const_reference back() const                    { return col_ptr.back(); }
};

inline void swap(Row &r1, Row &r2) { r1.swap(r2); }


enum ColumnType_t {
	ColumnType_integer =1,	//int64_t
	ColumnType_real    =2,	//double
	ColumnType_string  =3,	//asciiz
	ColumnType_raw     =4,	//binary
	ColumnType_boolean =5	//bool
};

class ColumnDescriptor {
public:
	ColumnType_t type;
	std::string name;
	
	ColumnDescriptor(ColumnType_t type_, const std::string &name_)
	  : type(type_),
	    name(name_)
	  {}
	bool operator==(const ColumnDescriptor &cd) const
	  { return type==cd.type && name==cd.name; }
	bool operator!=(const ColumnDescriptor &cd) const
	  { return type!=cd.type || name!=cd.name; }
};

typedef std::vector<ColumnDescriptor> RowSetDescriptor;

bool is_type_compatible(const RowSetDescriptor &desc1, const RowSetDescriptor &desc2);


class RowSet {
	RowSetDescriptor desc;
	typedef std::vector<Row> container_type;
	container_type row;
public:
	typedef Row value_type;
	typedef container_type::pointer pointer;
	typedef container_type::const_pointer const_pointer;
	typedef container_type::reference reference;
	typedef container_type::const_reference const_reference;
	typedef container_type::iterator iterator;
	typedef container_type::const_iterator const_iterator;
	typedef container_type::reverse_iterator reverse_iterator;
	typedef container_type::const_reverse_iterator const_reverse_iterator;
	typedef container_type::size_type size_type;
	typedef container_type::difference_type difference_type;
	
	
	RowSet() {}
	virtual ~RowSet();
	
	void setDescriptor(const RowSetDescriptor &new_desc);
	const RowSetDescriptor &descriptor() const { return desc; }
	
	void push_back(const Row &new_row);
	void add_and_swap(Row &new_row);
	
	void                    clear()               { row.clear(); }
	void                    reserve(size_type sz) { row.reserve(sz); }
	size_type               size() const          { return row.size(); }
	size_type               max_size() const      { return row.max_size(); }
	size_type               capacity() const      { return row.capacity(); }
	bool                    empty() const         { return row.empty(); }
	
	iterator                begin()               { return row.begin(); }
	const_iterator          begin() const         { return row.begin(); }
	iterator                end()                 { return row.end(); }
	const_iterator          end() const           { return row.end(); }
	reverse_iterator        rbegin()              { return row.rbegin(); }
	const_reverse_iterator  rbegin() const        { return row.rbegin(); }
	reverse_iterator        rend()                { return row.rend(); }
	const_reverse_iterator  rend() const          { return row.rend(); }
	reference       operator[](size_type n)       { return row[n]; }
	const_reference operator[](size_type n) const { return row[n]; }
	reference               front()               { return row.front(); }
	const_reference         front() const         { return row.front(); }
	reference               back()                { return row.back(); }
	const_reference         back() const          { return row.back(); }
};

#endif
