#define __STDC_FORMAT_MACROS
#include "ffdb_io.h"
#include "ffdb_op_alignment.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>


static bool parseDescriptorLine(char *line, RowSet *rowset) {
	RowSetDescriptor new_desc;
	bool first=true;
	char *ss;
	for(char *s=strtok_r(line," \t\r\n",&ss);
	    s;
	    s=strtok_r(0," \t\r\n",&ss))
	{
		if(first) {
			if(strcmp(s,"FFDB")!=0)
				return false;
			first=false;
		} else {
			char *p=strchr(s,':');
			if(!p) return false;
			*p='\0';
			p++;
			switch(*p) {
				case 'i':
					new_desc.push_back(ColumnDescriptor(ColumnType_integer,s));
					break;
				case 'f':
					new_desc.push_back(ColumnDescriptor(ColumnType_real,s));
					break;
				case 's':
					new_desc.push_back(ColumnDescriptor(ColumnType_string,s));
					break;
				case 'r':
					new_desc.push_back(ColumnDescriptor(ColumnType_raw,s));
					break;
				case 'b':
					new_desc.push_back(ColumnDescriptor(ColumnType_boolean,s));
					break;
				default:
					return false;
			}
		}
	}
	rowset->setDescriptor(new_desc);
	return true;
}


static inline void decodehex(void * dst_, size_t /*dstsize*/, const char * src, size_t srcsize) {
	//to be optimized
	uint8_t * dst = (uint8_t*)dst_;
	while(srcsize) {
		char c1 = *src++;
		char c2 = *src++;
		srcsize-=2;
		uint8_t xv1;
		if(c1>='0' && c1<='9') xv1=(uint8_t)(c1-'0');
		else if(c1>='a' && c1<='f') xv1=(uint8_t)(c1-'a'+10);
		else xv1=(uint8_t)(c1-'A'+10);
		uint8_t xv2;
		if(c2>='0' && c2<='9') xv2=(uint8_t)(c2-'0');
		else if(c2>='a' && c2<='f') xv2=(uint8_t)(c2-'a'+10);
		else xv2=(uint8_t)(c2-'A'+10);
		uint8_t value = (uint8_t)((xv1<<4)|xv2);
		*dst++ = value;
	}
}


#undef isdigit
#define isdigit(c) ((c)>='0' && (c)<='9')
#undef isspace
#define isspace(c) ((c)==' ' || (c)=='\t' || (c)=='\r' || (c)=='\n')
#undef isxdigit
#define isxdigit(c) (((c)>='0' && (c)<='9') || ((c)>='A' && (c)<='F') || ((c)>='a' && (c)<='f'))

//strtoll() would work too, but we do not need support for non-10 bases,
//thousands separator, etc. This is much faster. Really.
static inline int64_t strtoint64(const char *str, char **endptr) {
	bool is_negative=false;
	if(*str=='-') {
		str++;
		is_negative=true;
	}
	
	int64_t value=0;
	char c;
	while((c=*str)!='\0') {
		if(!isdigit(c))
			break;
		str++;
		value = value*10 + (c-'0');
	}
	*endptr = (char*)str;
	if(!is_negative)
		return value;
	else
		return -value;
}


static bool parseDataLine(char *line, size_t size_hint, RowSet *rowset) {
	const RowSetDescriptor &desc = rowset->descriptor();
	Row new_row;
	new_row.reserve_columns(desc.size());
	if(size_hint<4096) new_row.reserve_raw(size_hint);
	
	RowSetDescriptor::const_iterator desc_iter = desc.begin();
	while(*line) {
		while(isspace(*line))
			line++;
		if(!*line)
			break;

		if(line[0]=='?' && (line[1]=='\0' || isspace(line[1]))) {
			//NULL
			line++;
			new_row.alloc(0,0,0);
			desc_iter++;
			continue;
		}
		char *n=0;
		
		switch(desc_iter->type) {
			case ColumnType_integer: {
				int64_t value;
				char *endptr=0;
				value = strtoint64(line,&endptr);
				if(endptr) {
					n=endptr;
				} else {
					n=line;
					while(*n) n++;
				}
				
				void *ptr = new_row.alloc(sizeof(value),ffdb_type_alignment[ColumnType_integer]);
				*(int64_t*)ptr = value;
				
				break;
			}
			case ColumnType_real:{
				char *endptr=0;
				double value = strtod(line,&endptr);
				if(endptr)
					n=endptr;
				else {
					n=line;
					while(*n) n++;
				}
				void *ptr = new_row.alloc(sizeof(value),ffdb_type_alignment[ColumnType_real]);
				*(double*)ptr = value;
				
				break;
			}
			case ColumnType_string:{
				if(*line=='"') {
					std::string value;
					n = line+1;
					for(;;) {
						if(n[0]=='\0') break;
						if(n[0]=='"') break;
						if(n[0]=='\\') {
							n++;
						}
						value += n[0];
						n++;
					}
					if(n[0]=='"') n++;
					void *ptr = new_row.alloc(value.size()+1,ffdb_type_alignment[ColumnType_string],value.data());
					((char*)ptr)[value.size()] = '\0';
				} else {
					n=line;
					while(*n && !isspace(*n))
						n++;
					size_t sl = n-line;
					
					void *ptr = new_row.alloc(sl+1,ffdb_type_alignment[ColumnType_string],line);
					((char*)ptr)[sl] = '\0';
				}
				break;
			}
			case ColumnType_raw:{
				if(line[0]=='0' && line[1]=='x') line+=2;
				n=line;
				while(isxdigit(*n))
					n++;
				size_t hexsize = (size_t)(n-line);
				if(hexsize%2) return false;
				size_t rawsize = hexsize/2;
				void *ptr = new_row.alloc(rawsize,ffdb_type_alignment[ColumnType_raw]);
				decodehex(ptr,rawsize,line,hexsize);
				break;
			}
			case ColumnType_boolean: {
				n=line;
				while(*n && !isspace(*n))
					n++;
				size_t vl=(size_t)(n-line);
				bool value;
				if((vl==4 && strncmp(line,"true",4)==0) ||
				   (vl==1 && strncmp(line,"T",1)==0))
					value = true;
				else if((vl==5 && strncmp(line,"false",5)==0) ||
				        (vl==1 && strncmp(line,"F",1)==0))
					value = false;
				else
					return false;
				void *ptr = new_row.alloc(sizeof(value),ffdb_type_alignment[ColumnType_boolean]);
				*(bool*)ptr = value;
				break;
			}
		}
		if(*n && !isspace(*n))
			return false;
		desc_iter++;
		line = n;
	}
	if(new_row.size()!=desc.size())
		new_row.resize_columns(desc.size());
	size_hint = new_row.raw_size();
	
	rowset->add_and_swap(new_row);
	return true;
}


bool ReadFromFile(const char *filename, RowSet *rowset) {
	FILE *fp = fopen(filename,"r");
	if(!fp) return false;
	
	fseek(fp,0,SEEK_END);
	long file_size = ftell(fp);
	fseek(fp,0,SEEK_SET);
	
	setvbuf(fp,0,_IOFBF,65536);
	
	char line[4096];
	if(fgets(line,sizeof(line),fp)==0) {
		fclose(fp);
		return false;
	}
	
	if(!parseDescriptorLine(line,rowset)) {
		fclose(fp);
		return false;
	}
	
	size_t size_hint=0;
	int row_num=0;
	while(fgets(line,sizeof(line),fp)) {
		if(!parseDataLine(line,size_hint,rowset)) {
			fclose(fp);
			return false;
		}
		row_num++;
		if(row_num==1000) {
			double est_rows = file_size*1000.0/ftell(fp);
			if(est_rows>row_num)
				rowset->reserve((size_t)est_rows);
		}
	}
	fclose(fp);
//printf("%d rows read\n",(int)row_num);
	return true;
}




static bool writeDescriptor(FILE *fp, const RowSetDescriptor &desc) {
	fprintf(fp,"FFDB");
	for(RowSetDescriptor::const_iterator iter = desc.begin();
	    iter!=desc.end();
	    iter++)
	{
		fprintf(fp," %s:",iter->name.c_str());
		switch(iter->type) {
			case ColumnType_integer: putc('i',fp); break;
			case ColumnType_real:    putc('f',fp); break;
			case ColumnType_string:  putc('s',fp); break;
			case ColumnType_raw:     putc('r',fp); break;
			case ColumnType_boolean: putc('b',fp); break;
		}
	}
	putc('\n',fp);
	return true;
}


static bool writeRow(FILE *fp, const RowSetDescriptor &desc, const Row &row) {
//printf("row.size()=%d\n",row.size());
	Row::const_iterator rowcolumn_iter = row.begin();
	for(unsigned i=0; i<desc.size(); i++) {
		if(i!=0) putc(' ',fp);
		const void *value_ptr;
		if(rowcolumn_iter!=row.end())
			value_ptr = rowcolumn_iter->ptr;
		else
			value_ptr = 0;
//printf("  dumpign column %d:%s  ptr=%p\n",i,desc[i].name.c_str(),value_ptr);
		if(value_ptr) {
			switch(desc[i].type) {
				case ColumnType_integer: {
					int64_t value = *(const int64_t*)value_ptr;
					fprintf(fp,"%"PRId64, value);
					break;
				}
				case ColumnType_real: {
					double value = *(const double*)value_ptr;
					fprintf(fp,"%f", value);
					break;
				}
				case ColumnType_string: {
					const char *value=(const char *)value_ptr;
//printf("  sz=%d\n",rowcolumn_iter->size);
//printf("  v0=%c v1=%c v2=%c\n",value[0],value[1],value[2]);
					if(*value && strchr(value,' ')==0)
						fputs(value,fp);
					else {
						putc('"',fp);
						char c;
						while((c=*value++)!='\0') {
							if(c=='"' || c=='\\')
								putc('\\',fp);
							putc(c,fp);
						}
						putc('"',fp);
					}
					break;
				}
				case ColumnType_raw: {
					size_t sz=rowcolumn_iter->size;
					if(sz) { //to be optimized
						const uint8_t *value = (const uint8_t*)value_ptr;
						for(; sz; sz--) {
							uint8_t v = *value++;
							fprintf(fp,"%02"PRIx8, v);
						}
					} else
						fputs("0x",fp);
					break;
				}
				case ColumnType_boolean: {
					bool value = *(const bool*)value_ptr;
					fprintf(fp,value?"true":"false");
					break;
				}
			}
		} else {
			putc('?',fp);
		}
		
		if(rowcolumn_iter!=row.end())
			rowcolumn_iter++;
	}
	putc('\n',fp);
	return true;
}


bool WriteToFile(const char *filename, const RowSet &rowset) {
	FILE *fp = fopen(filename,"w");
	if(!fp) return false;
	
	setvbuf(fp,0,_IOFBF,65536);
	
	writeDescriptor(fp, rowset.descriptor());
	for(RowSet::const_iterator iter = rowset.begin();
	    iter!=rowset.end();
	    iter++)
	{
		writeRow(fp, rowset.descriptor(), *iter);
	}
	fclose(fp);
	return true;
}
