#ifndef _FFDB_OP_H_
#define _FFDB_OP_H_

#include "ffdb.h"
#include <vector>
#include <string>

#include "ffdb_op_rowsetdigest.h"

class DigestHelper {
public:
	typedef std::vector<unsigned> ColumnIdx;
	static void generate_rowset_digest(const RowSet &rowset, const ColumnIdx &column_idx, RowSetDigest *digest, std::vector<RowSet::const_pointer> *rowset_null_rows);
	static void generate_rowset_digest(const RowSet &rowset, const ColumnIdx &column_idx, RowSetDigest *digest);
	static void sort_rowset_digest(RowSetDigest *digest);
};


enum ConditionType_t {
	cond_null,
	cond_notnull,
	cond_lt,
	cond_le,
	cond_eq,
	cond_ge,
	cond_gt,
	cond_ne,
	cond_lt_col,
	cond_le_col,
	cond_eq_col,
	cond_ge_col,
	cond_gt_col,
	cond_ne_col
};



struct Cond_;
typedef bool (*cond_func_t_)(const Cond_ &cond, const Row::VL &vl, const Row &row);
struct Cond_ {
	std::string column_name;
	ConditionType_t condition_type;
	
	ColumnType_t value_type;
	std::string raw_value; //TODO: better data type

	ColumnType_t column_type;
	unsigned column_idx;
	
	cond_func_t_ cond_func;
};

class SelectionBase {
protected:
	std::vector<Cond_> condition;
	bool cache_valid;
	RowSetDescriptor cached_source_desc;
	bool prepare_cache_cond(const RowSetDescriptor &desc, Cond_ &cond);
	bool prepare_cache(const RowSet &row_set);
	
	bool match(const Row &row) const;
public:
	SelectionBase() : cache_valid(false) {}
	virtual ~SelectionBase();
	
	int add_condition(const char *column_name, ConditionType_t condition_type, const char *column_name2);
	int add_condition(const char *column_name, ConditionType_t condition_type);
	
	bool set_value(int condition_idx, ColumnType_t value_type, const void *value, size_t value_len);
	bool set_value(int condition_idx, int64_t value)
	  { return set_value(condition_idx,ColumnType_integer,&value,sizeof(value)); }
	bool set_value(int condition_idx, double value)
	  { return set_value(condition_idx,ColumnType_real,&value,sizeof(value)); }
	bool set_value(int condition_idx, const char *value);
	bool set_value(int condition_idx, const void *value, size_t valuelen)
	  { return set_value(condition_idx,ColumnType_raw,value,valuelen); }
	bool set_value(int condition_idx, bool value)
	  { return set_value(condition_idx,ColumnType_boolean,&value,sizeof(value)); }
};

class PreparedSelection : private SelectionBase {
public:
	SelectionBase::add_condition;
	SelectionBase::set_value;
	
	bool execute(const RowSet &row_set, Row *result);
	bool execute(const RowSet &row_set, RowSet *result);
};

class LockedSelection : private SelectionBase {
	const RowSet *rowset;
	RowSetDigest rowset_digest;
	RowSetDigest condition_digest;
	bool prepared;
	bool prepare();
public:
	LockedSelection() : rowset(0), prepared(false) {}
	~LockedSelection();
	
	SelectionBase::add_condition;
	SelectionBase::set_value;
	
	bool lock(const RowSet &rowset_);
	void unlock();
	
	bool execute(Row *result);
	bool execute(RowSet *result);
};



class PreparedProjection {
	std::vector<std::string> column_name;
	RowSetDescriptor cached_result_desc;
	RowSetDescriptor cached_source_desc;
	std::vector<unsigned> cached_col_idx;
	bool build_cache(const RowSet &row_set);
public:
	void add_column(const char *column_name);
	
	bool execute(const RowSet &row_set, RowSet *result);
};



class BinaryOpHelper {
public:
	
	std::vector<std::pair<std::string,std::string> > column_pair;
	DigestHelper::ColumnIdx cached_column1_idx;
	DigestHelper::ColumnIdx cached_column2_idx;
	bool prepare_cache(const RowSetDescriptor &desc1, const RowSetDescriptor &desc2);
	void add_relation(const char *column_name_1, const char *column_name_2);
	typedef std::vector<RowSet::const_pointer> RowIndexSet;
	
	struct Chunk {
		const RowSet *rowset;
		bool rowset_digest_valid;
		bool rowset_digest_sorted;
		RowSetDigest rowset_digest;
		bool rowset_null_rows_valid;
		RowIndexSet rowset_null_rows;
		Chunk() : rowset(0) {}
	};
	
	Chunk chunk1;
	Chunk chunk2;
	
	bool process_rowset(Chunk &chunk,
	                    const DigestHelper::ColumnIdx &column_idx,
			    bool sort, bool with_nulls);
	bool process_rowset_1(bool sort, bool with_nulls)
	  { return process_rowset(chunk1,cached_column1_idx,sort,with_nulls); }
	bool process_rowset_2(bool sort, bool with_nulls)
	  { return process_rowset(chunk2,cached_column2_idx,sort,with_nulls); }
	
	bool lock(Chunk &chunk, const RowSet &rowset);
	void unlock(Chunk &chunk);
	
	bool lock1(const RowSet &row_set_1)
	  { return lock(chunk1,row_set_1); }
	void unlock1() { unlock(chunk1); }
	bool lock2(const RowSet &row_set_2)
	  { return lock(chunk2,row_set_2); }
	void unlock2() { unlock(chunk1); }
	
	BinaryOpHelper() :
	  chunk1(), chunk2()
	  {}
	virtual ~BinaryOpHelper();
};


class LockedJoin : private BinaryOpHelper {
	bool cache_valid;
	RowSetDescriptor cached_desc1;
	RowSetDescriptor cached_desc2;
	RowSetDescriptor cached_desc_result;
	std::vector<size_t> cached_desc2_type_alignment;
	bool outer;
	
	bool prepare_cache(const RowSet &row_set_1, const RowSet &row_set_2);
	
	void join2rows(const Row &row1, const Row &row2, Row *result);
	
	void join_rows_with_nulls(RowSet *result);
	
	void execute_sort_merge(RowSet *result);
	void inner_loop_outer(RowSet *result);
	void inner_loop(RowSet *result);
public:
	LockedJoin() : BinaryOpHelper(), cache_valid(false) {}
	~LockedJoin();
	
	BinaryOpHelper::add_relation;
	BinaryOpHelper::lock1;
	BinaryOpHelper::unlock1;
	BinaryOpHelper::lock2;
	BinaryOpHelper::unlock2;
	
	void setOuter(bool outer_) { outer=outer_; }
	
	bool execute(RowSet *result);
};

class PreparedJoin : public LockedJoin {
public:
	bool execute(const RowSet &row_set_1, const RowSet &row_set_2, RowSet *result, bool outer=false);
};



class PreparedUnion {
public:
	bool execute(const RowSet &row_set_1, const RowSet &row_set_2, RowSet *result);
};


class PreparedUnique {
	std::vector<std::string> column_name;
public:
	void add_column(const char *column_name_);
	bool execute(const RowSet &rowset, RowSet *result);
};


class LockedIntersection : private BinaryOpHelper {
	bool cache_valid;
	RowSetDescriptor cached_desc1;
	RowSetDescriptor cached_desc2;
	bool prepare_cache(const RowSet &row_set_1, const RowSet &row_set_2);
public:
	LockedIntersection() : cache_valid(false) {}
	~LockedIntersection();
	BinaryOpHelper::add_relation;
	BinaryOpHelper::lock1;
	BinaryOpHelper::unlock1;
	BinaryOpHelper::lock2;
	BinaryOpHelper::unlock2;
	bool execute(RowSet *result);
};

class PreparedIntersection : public LockedIntersection {
public:
	bool execute(const RowSet &row_set_1, const RowSet &row_set_2, RowSet *result);
};


class LockedDifference : private BinaryOpHelper {
	bool cache_valid;
	RowSetDescriptor cached_desc1;
	RowSetDescriptor cached_desc2;
	bool prepare_cache(const RowSet &row_set_1, const RowSet &row_set_2);
public:
	LockedDifference() : cache_valid(false) {}
	~LockedDifference();
	BinaryOpHelper::add_relation;
	BinaryOpHelper::lock1;
	BinaryOpHelper::unlock1;
	BinaryOpHelper::lock2;
	BinaryOpHelper::unlock2;
	bool execute(RowSet *result);
};

class PreparedDifference : public LockedDifference {
public:
	bool execute(const RowSet &row_set_1, const RowSet &row_set_2, RowSet *result);
};


class LockedSubtract : private BinaryOpHelper {
	bool cache_valid;
	RowSetDescriptor cached_desc1;
	RowSetDescriptor cached_desc2;
	bool prepare_cache(const RowSet &row_set_1, const RowSet &row_set_2);
public:
	LockedSubtract() : cache_valid(false) {}
	~LockedSubtract();
	BinaryOpHelper::add_relation;
	BinaryOpHelper::lock1;
	BinaryOpHelper::unlock1;
	BinaryOpHelper::lock2;
	BinaryOpHelper::unlock2;
	bool execute(RowSet *result);
};

class PreparedSubtract : public LockedSubtract {
public:
	bool execute(const RowSet &row_set_1, const RowSet &row_set_2, RowSet *result);
};


//bool Selection(const RowSet &row_set, Row *result,
//               const char *column_name0, ...);

bool Projection(const RowSet &row_set, RowSet *result,
                const char *column_name0, ...);

bool Join(const RowSet &row_set_1, const RowSet &row_set_2, RowSet *result,
          const char *column_name_1_0, ...);

bool Union(const RowSet &row_set_1, const RowSet &row_set_2, RowSet *result);

bool Intersection(const RowSet &row_set_1, const RowSet &row_set_2, RowSet *result,
                  const char *column_name_1_0, ...);

bool Difference(const RowSet &row_set_1, const RowSet &row_set_2, RowSet *result,
                const char *column_name_1_0, ...);

bool Subtract(const RowSet &row_set_1, const RowSet &row_set_2, RowSet *result,
              const char *column_name_1_0, ...);

bool Unique(const RowSet &rowset, RowSet *result, const char *column_name_0, ...);

#endif
