#include "ffdb_op.h"
#include "ffdb_op_alignment.h"
#include "ffdb_op_cmpdigest.h"
#include <stdarg.h>
#include <string.h>
#include <algorithm>
#include <assert.h>

#include <stdio.h>

static bool cond_null_func(const Cond_ &/*cond*/, const Row::VL &vl, const Row &/*row*/) {
	if(vl.ptr) return false;
	return true;
}

static bool cond_notnull_func(const Cond_ &/*cond*/, const Row::VL &vl, const Row &/*row*/) {
	if(!vl.ptr) return false;
	return true;
}


static int cmp(size_t sz1, const void *ptr1, ColumnType_t value_type, size_t sz2, const void *ptr2) {
	switch(value_type) {
		case ColumnType_integer: {
			assert(sz1==sizeof(int64_t));
			assert(sz2==sizeof(int64_t));
			int64_t value1 = *(int64_t*)ptr1;
			int64_t value2 = *(int64_t*)ptr2;
			if(value1<value2) return -1;
			if(value1>value2) return 1;
			return 0;
		}
		case ColumnType_real: {
			assert(sz1==sizeof(double));
			assert(sz2==sizeof(double));
			double value1 = *(double*)ptr1;
			double value2 = *(double*)ptr2;
			if(value1<value2) return -1;
			if(value1>value2) return 1;
			return 0;
		}
		case ColumnType_string: {
			size_t min_sz=sz1;
			if(sz2<min_sz) min_sz=sz2;
			int rc=strncmp((const char*)ptr1, (const char*)ptr2, min_sz);
			if(rc<0) return -1;
			if(rc>0) return 1;
			if(sz1<sz2) return -1;
			if(sz2>sz2) return 1;
			return 0;
		}
		case ColumnType_raw: {
			if(sz1!=sz2) return -1; //not equal
			if(memcmp(ptr1,ptr2,sz1)!=0) return -1;
			return 0;
		}
		case ColumnType_boolean: {
			assert(sz1==sizeof(bool));
			assert(sz2==sizeof(bool));
			bool value1 = *(bool*)ptr1;
			bool value2 = *(bool*)ptr2;
			if(value1<value2) return -1;
			if(value1>value2) return 1;
			return 0;
		}
	}
	//NOTREACHED
	return -1;
}

static int cmp(const Row::VL &vl, ColumnType_t value_type, const std::string &raw_value) {
	return cmp(vl.size,vl.ptr,value_type,raw_value.size(),raw_value.data());
}

static int cmp(const Row::VL &vl1, ColumnType_t value_type, const Row::VL &vl2) {
	return cmp(vl1.size,vl1.ptr,value_type,vl1.size,vl2.ptr);
}


static bool cond_lt_func(const Cond_ &cond, const Row::VL &vl, const Row &/*row*/) {
	if(vl.ptr==0) return false;
	return cmp(vl,cond.value_type,cond.raw_value) < 0;
}
static bool cond_le_func(const Cond_ &cond, const Row::VL &vl, const Row &/*row*/) {
	if(vl.ptr==0) return false;
	return cmp(vl,cond.value_type,cond.raw_value) <= 0;
}
static bool cond_eq_func(const Cond_ &cond, const Row::VL &vl, const Row &/*row*/) {
	if(vl.ptr==0) return false;
	return cmp(vl,cond.value_type,cond.raw_value) == 0;
}
static bool cond_ge_func(const Cond_ &cond, const Row::VL &vl, const Row &/*row*/) {
	if(vl.ptr==0) return false;
	return cmp(vl,cond.value_type,cond.raw_value) >= 0;
}
static bool cond_gt_func(const Cond_ &cond, const Row::VL &vl, const Row &/*row*/) {
	if(vl.ptr==0) return false;
	return cmp(vl,cond.value_type,cond.raw_value) > 0;
}
static bool cond_ne_func(const Cond_ &cond, const Row::VL &vl, const Row &/*row*/) {
	if(vl.ptr==0) return false;
	return cmp(vl,cond.value_type,cond.raw_value) != 0;
}

static bool cond_lt_col_func(const Cond_ &cond, const Row::VL &vl, const Row &row) {
	if(vl.ptr==0) return false;
	unsigned other_column_idx = *(unsigned*)cond.raw_value.data();
	const Row::VL &vl2 = row[other_column_idx];
	if(vl2.ptr==0) return false;
	return cmp(vl,cond.column_type,vl2) < 0;
}
static bool cond_le_col_func(const Cond_ &cond, const Row::VL &vl, const Row &row) {
	if(vl.ptr==0) return false;
	unsigned other_column_idx = *(unsigned*)cond.raw_value.data();
	const Row::VL &vl2 = row[other_column_idx];
	if(vl2.ptr==0) return false;
	return cmp(vl,cond.column_type,vl2) <= 0;
}
static bool cond_eq_col_func(const Cond_ &cond, const Row::VL &vl, const Row &row) {
	if(vl.ptr==0) return false;
	unsigned other_column_idx = *(unsigned*)cond.raw_value.data();
	const Row::VL &vl2 = row[other_column_idx];
	if(vl2.ptr==0) return false;
	return cmp(vl,cond.column_type,vl2) == 0;
}
static bool cond_ge_col_func(const Cond_ &cond, const Row::VL &vl, const Row &row) {
	if(vl.ptr==0) return false;
	unsigned other_column_idx = *(unsigned*)cond.raw_value.data();
	const Row::VL &vl2 = row[other_column_idx];
	if(vl2.ptr==0) return false;
	return cmp(vl,cond.column_type,vl2) >= 0;
}
static bool cond_gt_col_func(const Cond_ &cond, const Row::VL &vl, const Row &row) {
	if(vl.ptr==0) return false;
	unsigned other_column_idx = *(unsigned*)cond.raw_value.data();
	const Row::VL &vl2 = row[other_column_idx];
	if(vl2.ptr==0) return false;
	return cmp(vl,cond.column_type,vl2) > 0;
}
static bool cond_ne_col_func(const Cond_ &cond, const Row::VL &vl, const Row &row) {
	if(vl.ptr==0) return false;
	unsigned other_column_idx = *(unsigned*)cond.raw_value.data();
	const Row::VL &vl2 = row[other_column_idx];
	if(vl2.ptr==0) return false;
	return cmp(vl,cond.column_type,vl2) != 0;
}



static bool cond_integer_eq_func(const Cond_ &cond, const Row::VL &vl, const Row &/*row*/) {
	if(vl.ptr==0) return false;
	return *(const int64_t*)vl.ptr == *(const int64_t*)cond.raw_value.data();
}

static bool cond_string_eq_func(const Cond_ &cond, const Row::VL &vl, const Row &/*row*/) {
	if(vl.ptr==0) return false;
	if(vl.size!=cond.raw_value.size()) return false;
	return memcmp(vl.ptr,cond.raw_value.data(),vl.size)==0;
}


bool SelectionBase::prepare_cache_cond(const RowSetDescriptor &desc, Cond_ &cond) {
	unsigned column_idx;
	for(column_idx=0; column_idx<desc.size(); column_idx++)
		if(desc[column_idx].name==cond.column_name) break;
	if(column_idx>=desc.size())
		return false; //unknown column
	
	cond.column_type = desc[column_idx].type;
	cond.column_idx = column_idx;
	
	if(cond.condition_type!=cond_null && cond.condition_type!=cond_notnull) {
		if(cond.condition_type>=cond_lt_col && cond.condition_type<=cond_ne_col) {
			if(cond.raw_value.size()==0) return false;
			std::string other_column_name;
			if(cond.value_type==ColumnType_string) {
				//stored value is the column name
				other_column_name = cond.raw_value;
			} else if(cond.value_type==(ColumnType_t)0) {
				other_column_name = cond.raw_value.substr(sizeof(unsigned));
			} else
				return false;
			unsigned other_column_idx;
			for(other_column_idx=0; other_column_idx<desc.size(); other_column_idx++)
				if(desc[other_column_idx].name==other_column_name) break;
			if(other_column_idx>=desc.size())
				return false; //unknown column
			if(cond.column_type!=desc[other_column_idx].type) return false;
			cond.raw_value.resize(sizeof(unsigned)+other_column_name.size());
			memcpy(cond.raw_value.begin()+0, &other_column_idx, sizeof(unsigned));
			memcpy(cond.raw_value.begin()+sizeof(unsigned), other_column_name.data(), other_column_name.size());
			cond.value_type = (ColumnType_t)0;
		} else if(cond.column_type != cond.value_type)
			return false; //incompatible
			//todo: type coercion
	}
	
	static const cond_func_t_ cond_func[]={
		cond_null_func,
		cond_notnull_func,
		cond_lt_func,
		cond_le_func,
		cond_eq_func,
		cond_ge_func,
		cond_gt_func,
		cond_ne_func,
		cond_lt_col_func,
		cond_le_col_func,
		cond_eq_col_func,
		cond_ge_col_func,
		cond_gt_col_func,
		cond_ne_col_func
	};
	
	cond.cond_func = cond_func[cond.condition_type];
	//optimized: (does not really gain that much...)
	if(cond.condition_type==cond_eq && cond.column_type==ColumnType_integer)
		cond.cond_func = cond_integer_eq_func;
	if(cond.condition_type==cond_eq && cond.column_type==ColumnType_string)
		cond.cond_func = cond_string_eq_func;
	return true;
}


bool SelectionBase::prepare_cache(const RowSet &row_set) {
	const RowSetDescriptor &desc = row_set.descriptor();
	if(cache_valid && desc==cached_source_desc) return true;
	cache_valid = false;
	
	for(unsigned i=0; i<condition.size(); i++) {
		if(!prepare_cache_cond(desc,condition[i]))
			return false;
	}
	cached_source_desc = desc;
	cache_valid = true;
	return true;
}





int SelectionBase::add_condition(const char *column_name, ConditionType_t condition_type, const char *column_name2) {
	if(condition_type>=cond_lt || condition_type<=cond_ne)
		condition_type = (ConditionType_t)(condition_type + 6);
	if(condition_type<cond_lt_col || condition_type>cond_ne_col)
		return -1;
	Cond_ cond;
	cond.column_name = column_name;
	cond.condition_type = condition_type;
	cond.value_type = ColumnType_string;
	cond.raw_value.assign(column_name2,strlen(column_name2));
	cond.column_type = (ColumnType_t)0;
	
	condition.push_back(cond);
	cache_valid = false;
	return (int)condition.size()-1;
}


int SelectionBase::add_condition(const char *column_name, ConditionType_t condition_type) {
	if(condition_type>=cond_lt_col && condition_type<=cond_ne_col)
		return -1;
	
	Cond_ cond;
	cond.column_name = column_name;
	cond.condition_type = condition_type;
	cond.value_type = (ColumnType_t)0; //yes, it is a bit naughty
	cond.column_type = (ColumnType_t)0;
	
	condition.push_back(cond);
	cache_valid = false;
	return (int)condition.size()-1;
}



bool SelectionBase::set_value(int condition_idx, ColumnType_t value_type, const void *value, size_t value_len)
{
	if(condition_idx<0 || (unsigned)condition_idx>=condition.size())
		return false;
	Cond_ &cond = condition[condition_idx];
	if(cond.condition_type==cond_null ||
	   cond.condition_type==cond_notnull ||
	  (cond.condition_type>=cond_lt_col && cond.condition_type<=cond_ne_col))
		return false; //cannot specify value for those conditions
	if(value_type==ColumnType_raw &&
	   cond.condition_type!=cond_eq &&
	   cond.condition_type!=cond_ne)
		return false; //not supported for raw
	
	if(cond.value_type!=value_type) cache_valid = false;

	cond.value_type = value_type;
	cond.raw_value.assign((const char*)value,value_len);
	return true;
}


bool SelectionBase::set_value(int condition_idx, const char *value) {
	return set_value(condition_idx, ColumnType_string, value, strlen(value)+1);
}


SelectionBase::~SelectionBase() {
}



bool SelectionBase::match(const Row &row) const {
	for(unsigned i=0; i<condition.size(); i++) {
		const Cond_ &cond = condition[i];
		if(!cond.cond_func(cond,row[cond.column_idx],row))
			return false;
	}
	return true;
}

bool PreparedSelection::execute(const RowSet &row_set, Row *result) {
	if(!prepare_cache(row_set)) return false;
	for(RowSet::const_iterator iter=row_set.begin();
	    iter!=row_set.end();
	    iter++)
	{
		if(match(*iter)) {
			*result = *iter;
			return true;
		}
	}
	return false;
}

bool PreparedSelection::execute(const RowSet &row_set, RowSet *result) {
	if(!prepare_cache(row_set)) return false;
	
	result->setDescriptor(row_set.descriptor());
	for(RowSet::const_iterator iter=row_set.begin();
	    iter!=row_set.end();
	    iter++)
	{
		if(match(*iter))
			result->push_back(*iter);
	}
	return true;
}


bool LockedSelection::prepare() {
	if(prepared) return true;
	if(!SelectionBase::prepare_cache(*rowset)) return false;
	
	//Transform eq-conditions into column idx
	DigestHelper::ColumnIdx column_idx;
	for(std::vector<Cond_>::const_iterator iter = condition.begin();
	    iter!=condition.end();
	    iter++)
	{
//printf("looking at condition #%u\n",iter-condition.begin());
		if(iter->condition_type==cond_eq) {
//printf(" Is cond_eq\n");
			column_idx.push_back(iter->column_idx);
		}
	}
	//generate rowset digest
	DigestHelper::generate_rowset_digest(*rowset,column_idx,&rowset_digest);
	DigestHelper::sort_rowset_digest(&rowset_digest);
	
	
	//Transform eq-condition-values into rowdigest
	//transform values into row
//printf("transforming conditions into a set digest\n");
	RowSetDescriptor tmp_desc;
	Row tmp_row;
	unsigned ci=0;
	for(RowSetDescriptor::const_iterator desc_iter = cached_source_desc.begin();
	    desc_iter != cached_source_desc.end();
	    desc_iter++,ci++)
	{
		bool has_eq_cond=false;
//printf("  doing column %u\n",ci);
		for(std::vector<Cond_>::const_iterator cond_iter = condition.begin();
		    cond_iter!=condition.end();
		    cond_iter++)
		{
//printf("    looking at cond #%u\n", cond_iter-condition.begin());
			const Cond_ &cond = *cond_iter;
			if(cond.column_idx!=ci) continue;
//printf("    ->has same column_idx\n");
			if(cond.condition_type==cond_eq) {
//printf("    ->is cond_eq\n");
				tmp_desc.push_back(ColumnDescriptor(cond.value_type,""));
				tmp_row.alloc(cond.raw_value.size(),ffdb_type_alignment[cond.value_type],cond.raw_value.data());
				has_eq_cond=true;
				break;
			}
		}
		if(!has_eq_cond) {
//printf("    ->column %u does not have associated cond_eq\n",ci);
			tmp_row.alloc(0,0,0);
		}
	}
//printf("done transforming\n");
	//transform into rowset
	RowSet tmp_rowset;
	tmp_rowset.setDescriptor(tmp_desc);
	tmp_rowset.push_back(tmp_row);
	//transform into rowsetdigest
	DigestHelper::generate_rowset_digest(tmp_rowset,column_idx,&condition_digest);
	
	prepared = true;
	
	return true;
}


bool LockedSelection::lock(const RowSet &rowset_) {
	unlock();
	rowset = &rowset_;
	return true;
}

void LockedSelection::unlock() {
	rowset = 0;
	rowset_digest.clear();
	prepared = false;
}
	

bool LockedSelection::execute(Row *result) {
	if(!rowset) return false;
	if(!prepare()) return false;
	
	Compare_RowDigest cmp;
	RowSetDigest::const_iterator iter = std::lower_bound(rowset_digest.begin(),
	                                                     rowset_digest.end(),
	                                                     *(condition_digest.begin()),
	                                                     cmp
	                                                    );
	RowSetDigest::const_iterator last = std::upper_bound(rowset_digest.begin(),
	                                                     rowset_digest.end(),
	                                                     *(condition_digest.begin()),
	                                                     cmp
	                                                    );
	while(iter!=last) {
		const Row &row = *iter->row_ptr;
		if(match(row)) {
			*result = row;
			return true;
		}
		iter++;
	}
	return false;
}

bool LockedSelection::execute(RowSet *result) {
	if(!rowset) return false;
	if(!prepare()) return false;
	
	result->setDescriptor(rowset->descriptor());
	
	Compare_RowDigest cmp;
	std::pair<RowSetDigest::const_iterator,RowSetDigest::const_iterator> ip =
	std::equal_range(((const RowSetDigest&)rowset_digest).begin(),
	                 ((const RowSetDigest&)rowset_digest).end(),
	                 *(condition_digest.begin()),
	                 cmp);
	RowSetDigest::const_iterator iter=ip.first;
	RowSetDigest::const_iterator last=ip.second;
//printf("lower offset=%u\n",iter-rowset_digest.begin());
//printf("upper offset=%u\n",last-rowset_digest.begin());
	while(iter!=last) {
		const Row &row = *iter->row_ptr;
//printf("iter->row_index=%u\n",iter->row_index);
		if(match(row))
			result->push_back(row);
		iter++;
	}
	return true;
}

LockedSelection::~LockedSelection() {
}
